// MapPanel.java
// Started on June 16th, 2001
// Copyright (c) Virtu Halttunen

//import TheGame.*;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
import java.util.Vector;

public class MapPanel extends JPanel implements Scrollable
{
    public final static int LIGHT_LEVELS = 32;
    public final static int SELECTED = 1;

    private Map map;
    private Dimension viewportSize;
    private int cellSize;
    private Vector floorCells;
    private Vector wallCells;
    private Image[][] doorCells;
    private Vector objectCells;
    private int openDoorIndex=0;
    private int closedDoorIndex=0;
    private Color [] lights;
    private Color highlightColor = Color.yellow;
    private boolean drawLights = true;
    private boolean drawRooms = false;
    private boolean drawObjectBounds = false;
    private Image alertIcon;
    
    public MapPanel( Map map, int cellSize, int visibleHorizontal, int visibleVertical )
    {
	super();
	this.map = map;
	this.cellSize = cellSize;
	viewportSize = new Dimension( visibleHorizontal*cellSize, visibleVertical*cellSize );
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
	loadFloorsFromMap();
	loadWallsFromMap();
	loadDoorsFromMap();
	loadObjectsFromMap();
	lights = new Color[ LIGHT_LEVELS ];
	for( int i = 0; i < LIGHT_LEVELS; i++ )
	{
	    lights[ i ] = new Color( 1, 1, 1, i*(255/LIGHT_LEVELS));
	}
	//setCursor( Cursor.getPredefinedCursor( Cursor.CROSSHAIR_CURSOR ) );
	alertIcon = ResourceManager.loadImage( "AlertIcon.png" );
	setAutoscrolls( true );
    }
    
    protected void paintComponent( Graphics g )
    {
	super.paintComponent( g );
	//System.out.println( "MapPanel: repaint" );
	Rectangle newRect = g.getClipBounds();
	Point min = getCell( new Point( newRect.x, newRect.y ) );
	int minX = min.x;
	int minY = min.y;
	Point max = getCell( new Point( newRect.x+newRect.width, newRect.y+newRect.height ) );
	int maxX = max.x;
	int maxY = max.y;
	
	if( maxX == map.getHorizontalSize() ) maxX = maxX-1;
	if( maxY == map.getVerticalSize() ) maxY = maxY-1;
	  
	for( int i = minX; i <= maxX; i++ )
	{
	    for( int j = minY; j <= maxY; j++ )
	    {
		drawCell( g, i, j );
		if( map.getUtility( i, j ) == MapMaker.SELECTED )
		{
		    g.setColor( highlightColor );
		    g.drawRect( i*cellSize+1, j*cellSize+1, cellSize-2, cellSize-2 );
		    g.drawRect( i*cellSize+2, j*cellSize+2, cellSize-4, cellSize-4 );
		}
	    }	
	}
	// Paint objects
	if( drawObjectBounds )
	{
	    for( int i=0; i < map.getNumberOfObjects(); i++ )
	    {
		MapObject mapObject = map.getObject( i );
		g.setColor( Color.white );
		minX = mapObject.getTopLeft().x*cellSize;
		minY = mapObject.getTopLeft().y*cellSize;
		maxX = mapObject.getBottomRight().x*cellSize+cellSize;
		maxY = mapObject.getBottomRight().y*cellSize+cellSize;
		g.drawRect( minX, minY, maxX-minX, maxY-minY );
		g.drawString( mapObject.getName(), minX+3, minY+15 );
	    }
	}
	// Paint rooms
	if( drawRooms )
	{
	    Room [] rooms = map.getRooms();
	    Room room;
	    if( rooms != null && rooms.length > 0 )
	    {
		for( int i=0; i < rooms.length; i++ )
		{
		    g.setColor( Color.white );
		    room = rooms[i];
		    minX = room.getTopLeft().x*cellSize;
		    minY = room.getTopLeft().y*cellSize;
		    maxX = room.getBottomRight().x*cellSize+cellSize;
		    maxY = room.getBottomRight().y*cellSize+cellSize;
		    g.drawRect( minX, minY, maxX-minX, maxY-minY );
		    g.drawString( room.getName(), minX+15, minY+15 );
		    Point hotSpot = room.getHotSpot();
		    g.setColor( Color.cyan );
		    String [] linkedIds = room.getLinkedRoomIds();
		    for( int j=0; j < linkedIds.length; j++ )
		    {
			Room otherRoom = map.getRoom( linkedIds[ j ] );
			if( otherRoom != null )
			{
			    Point otherHotSpot = otherRoom.getHotSpot();
			    g.drawLine( hotSpot.x*cellSize+cellSize/2, hotSpot.y*cellSize+cellSize/2, otherHotSpot.x*cellSize+cellSize/2, otherHotSpot.y*cellSize+cellSize/2 );
			}
		    }
		    g.setColor( Color.white );
		    g.fillOval( hotSpot.x*cellSize+10, hotSpot.y*cellSize+10, cellSize-20, cellSize-20 );

		}
	    }
	}
	
    }
    
    private void drawCell( Graphics g, int x, int y )
    {
	MapCell cell =  map.getCell( x, y );
        g.setColor( Color.black );
        g.fillRect( x*cellSize, y*cellSize, cellSize, cellSize );

	if( cell != null )
        {
            int floor;
            for( int i = 0; i < cell.getFloorSize(); i++ )
            {
                floor = cell.getFloor( i );
                if( floor >= 0 )
                    g.drawImage( (Image)floorCells.elementAt( floor ), x*cellSize, y*cellSize, cellSize, cellSize, this );
            }
            if( cell.hasObjects() )
            {
		// Draw objects
		for( int i=0; i < map.getNumberOfObjects(); i++ )
		{
		    MapObject mapObject = map.getObject( i );
		    if( mapObject.isLocatedIn( x, y ) )
		    {
			MapObjectCell[] mapObjectCells = mapObject.getCells( x, y );
			if( mapObjectCells != null )
			{
			    for( int j=0; j < mapObjectCells.length; j++ )
			    {
				if( mapObjectCells[j].isLocatedIn( x, y ) )
				{
				    int objIndex = mapObjectCells[j].getType();
				    if( mapObjectCells[j].isDestroyed() )
				    {
					objIndex = mapObjectCells[j].getDestroyedType();
					if( objIndex < 0 )
					{
					    objIndex = mapObjectCells[j].getType();
					}					
				    }
				    g.drawImage( (Image)objectCells.elementAt( objIndex ), x*cellSize, y*cellSize, cellSize, cellSize, this );
				}
			    }
			}
		    }
		}
		// Draw unbound object cells
                for( int i=0; i < map.getNumberOfUnboundObjects(); i++ )
                {
                    MapObjectCell mapObjectCell = map.getObjectCell( i );
                    if( mapObjectCell.isLocatedIn( x, y ) )
                    {
                        int objIndex = mapObjectCell.getType();
                        if( mapObjectCell.isDestroyed() )
			{
			    objIndex = mapObjectCell.getDestroyedType();
			    if( objIndex < 0 )
			    {
				objIndex = mapObjectCell.getType();
			    }
			}
                        g.drawImage( (Image)objectCells.elementAt( objIndex ), x*cellSize, y*cellSize, cellSize, cellSize, this );
			g.drawImage( alertIcon, x*cellSize, y*cellSize, alertIcon.getWidth( this ), alertIcon.getHeight( this ), this );
                    }
                }
            }
            int wall = cell.getWall();
            if( wall >= 0 )
            {
                g.drawImage( (Image)wallCells.elementAt( wall ), x*cellSize, y*cellSize, cellSize, cellSize, this );
            }
            else
            {
                // If it's a door
                if( wall < -1 )
                {
                    Door door = map.getDoor( x, y );
                    if( door != null )
                    {
                        int dir = door.getDirection();
                        int dIndex = 0;
                        int doorState;
                        if( dir == Map.NORTH ) dIndex = 1;
                        if( dir == Map.SOUTH ) dIndex = 0;
                        if( dir == Map.EAST ) dIndex = 2;
                        if( dir == Map.WEST ) dIndex = 3;
                        // Draw the door
                        if( door.isOpen() )
                            doorState = openDoorIndex;
                        else
                            doorState = closedDoorIndex;
                        g.drawImage( doorCells[ dIndex ][ doorState ], x*cellSize, y*cellSize, cellSize, cellSize, this );
                    }
                }
            }
            if( drawLights )
            {
                g.setColor( lights[ cell.getLight() ] );
                g.fillRect( x*cellSize, y*cellSize, cellSize, cellSize );
            }
        }
	if( map.getUtility( x, y ) == SELECTED )
	{
	    g.setColor( highlightColor );
	    g.drawRect( x*cellSize, y*cellSize, cellSize, cellSize );
	    g.drawRect( x*cellSize+1, y*cellSize+1, cellSize-2, cellSize-2 );
	}
    }

        /*
    public Image selectImageForCellType( int cellType )
    {
	if( cellType < -Map.DOORS )
	{
	   return (Image)floorCells.elementAt( (-cellType)-(Map.DOORS+1) );
	} 
	else if( cellType > Map.DOORS && cellType < Map.OBJECTS )
	{
	   return (Image)wallCells.elementAt( cellType-(Map.DOORS+1) );
	}
	else if( cellType > -Map.DOORS && cellType < 0 )
	{
	    return doorCells[ Math.abs( cellType )-1 ][ openDoorIndex ];
	}
	else if( cellType > 0 && cellType < Map.DOORS )
	{	
	    return doorCells[ cellType-1 ][ closedDoorIndex ];
	}
	else if( cellType > Map.OBJECTS )
	{
	    return (Image)objectCells.elementAt( cellType-(Map.OBJECTS+1) );
	} 
	else
	    return null;
    }*/
    
    public int getCellSize()
    {
	return cellSize;
    }
    
    public void setCellSize( int cellSize )
    {
	this.cellSize = cellSize;
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	invalidate();
	repaint();
    }
    
    public void setDrawLights( boolean drawLights )
    {
	this.drawLights = drawLights;
	repaint();
    }
    
    public void setDrawRooms( boolean drawRooms )
    {
	this.drawRooms = drawRooms;
    }
    
    public void setDrawObjectBounds( boolean drawObjectBounds )
    {
	this.drawObjectBounds = drawObjectBounds;
    }
    
    public Point getCell( Point point )
    {
	int x=0;
	int y=0;
	
	x = point.x/cellSize;
	y = point.y/cellSize;
	if( x < 0 ) x = 0;
	if( x > map.getHorizontalSize() ) x = map.getHorizontalSize()-1;
	if( y < 0 ) y = 0;
	if( y > map.getVerticalSize() ) y = map.getVerticalSize()-1;
	
	return new Point( x, y );
    }

    private void loadFloorsFromMap()
    {
	Vector floorVector = map.getFloorSets();
	if( floorVector.size() > 0 )
	{
	    for( int i = 0; i < floorVector.size(); i++ )
	    {
		//loadSet( (String)floorVector.elementAt( i ), floorCells );
		ResourceManager.cropImage( (String)floorVector.elementAt( i ), floorCells, cellSize );
	    }
	}	
    }

    private void loadWallsFromMap()
    {
	Vector wallVector = map.getWallSets();
	if( wallVector.size() > 0 )
	{
	    for( int i = 0; i < wallVector.size(); i++ )
	    {
		//loadSet( (String)wallVector.elementAt( i ), wallCells );
		ResourceManager.cropImage( (String)wallVector.elementAt( i ), wallCells, cellSize );
	    }
	}	
    }

    private void loadDoorsFromMap()
    {
	System.out.println( "Loading door sets" );
	Vector doorVector = map.getDoorSets();
	Vector doorImages = new Vector();
	if( doorVector.size() > 0 )
	{
	    for( int i = 0; i < doorVector.size(); i++ )
	    {
		ResourceManager.cropImage( (String)doorVector.elementAt( i ), doorImages, cellSize );
	    }
	}
	doorCells = new Image[4][doorImages.size()/4];
	System.out.println( doorImages.size()/4 + " door frame(s) found" );
	openDoorIndex = doorImages.size()/4-1;
	closedDoorIndex = 0;
	for( int i=0; i < 4; i++ )
	{
	    for( int j=0; j < doorImages.size()/4;j++ )
	    {
		doorCells[i][j] = (Image)doorImages.elementAt( i+j*4 );
		System.out.println( "Door image ("+i+","+j+") set from "+(i+j*4) );
	    }
	}	
    }

    public void loadObjectsFromMap()
    {
	objectCells = new Vector();
	Vector objectVector = map.getObjectSets();
	if( objectVector.size() > 0 )
	{
	    for( int i = 0; i < objectVector.size(); i++ )
	    {
		//loadSet( (String)objectVector.elementAt( i ), objectCells );
		ResourceManager.cropImage( (String)objectVector.elementAt( i ), objectCells, cellSize );
	    }
	}	
    }

/*
    private void loadSet( String setFileName, Vector vector )
    {
	ImageFilter filter;
	ImageProducer producer;
	Image theSetImage = new ImageIcon( getClass().getResource( setFileName ) ).getImage(); // Uses ImageIcon to really load the image instead of waiting for the first repaint
	if( ( theSetImage.getWidth( this ) % cellSize == 0 ) && ( theSetImage.getHeight( this) % cellSize == 0 ) )
	{
	    for( int i = 0; i < theSetImage.getHeight( this )/cellSize; i++ )
		for( int j = 0; j < theSetImage.getWidth( this )/cellSize; j++ )
		{
		    filter = new CropImageFilter( j*cellSize, i*cellSize, cellSize, cellSize );
		    producer = new FilteredImageSource( theSetImage.getSource(), filter );
		    vector.addElement( createImage( producer ) );
		}
	}
	System.out.println( "MapPanel: Loaded set " + setFileName +", " + vector.size() + " images found" );
    }*/
    
    public void changeMap( Map map )
    {
	this.map = map;
	Dimension mySize = new Dimension ( map.getHorizontalSize()*cellSize, map.getVerticalSize()*cellSize );
	setMinimumSize( mySize );
	setPreferredSize( mySize );
	setMaximumSize( mySize );
	floorCells = new Vector();
	wallCells = new Vector();
	objectCells = new Vector();
	loadFloorsFromMap();
	loadWallsFromMap();
	loadDoorsFromMap();
	loadObjectsFromMap();
	repaint();
    }
    
    public Vector getFloorCells()
    {
	return floorCells;
    }
    
    public Vector getWallCells()
    {
	return wallCells;
    }
    
    public Vector getObjectCells()
    {
	return objectCells;
    }
    
    public Image[][] getDoorCells()
    {
	return doorCells;
    }
    
    public Dimension getPreferredScrollableViewportSize()
    {
	return viewportSize;
    }
    
    public int getScrollableUnitIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	//get the current position
	int currentPosition = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	    currentPosition = visibleRect.x;
	else
	    currentPosition = visibleRect.y;
    
	//return the number of pixels between currentPosition
	//and the nearest tick mark in the indicated direction
	if ( direction < 0 ) 
	{
	    int newPosition = currentPosition - (currentPosition / cellSize) * cellSize;
	    return (newPosition == 0) ? cellSize : newPosition;
	} 
	else 
	{
	    return ((currentPosition / cellSize) + 1) * cellSize - currentPosition;
	}
    }
    
    public int getScrollableBlockIncrement( Rectangle visibleRect, int orientation, int direction ) 
    {
	int amount = 0;
	if( orientation == SwingConstants.HORIZONTAL )
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.width - amount;
	}
	else
	{
	    amount = getScrollableUnitIncrement( visibleRect, orientation, direction );
	    if( amount < cellSize )
		amount = amount+cellSize;
	    return visibleRect.height - amount;
	}
    }
    
    public boolean getScrollableTracksViewportWidth()
    {
	return false;
    }
    
    public boolean getScrollableTracksViewportHeight()
    {
	return false;
    }

    
}