//
// MissionEditor.java
// Started on July 21st, 2001
// Copyright (c) Virtu Halttunen

// April 12th, 2002: Major rewrite to support the new map format. Also laid the ground for adding monsters and
// Objectives. Some hackish dummy classes were required
// May 19th, 2002: The GUI for adding and modifying monsters and Objectives starts to be in place. Got
// rid of the dummy classes and added a ton of references from The Game instead. Sigh. Really would need a
// better way.
// May 22nd, 2002: Added a panel for editing monsters and player characters. The GUI is almost done...
// May 23rd, 2002: The GUI is done, characters can be rotated, and random bounds can be set for character
// classes
// May 25th, 2002: Added fields for setting mission descriptions
// May 26th, 2002: The missions can be saved and loaded
// September 2nd, 2003: Graphics, objectives and monsters can now be loaded straight from theGame.jar
// July 23rd, 2005: Started adding support for the new XML-based features
// July 25th, 2005: Lots of interface tweaks

import TheGame.Objective;
import TheGame.MapCell;
import TheGame.CharacterSprite;
import TheGame.GameAction;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;
import java.util.zip.*;
import java.io.*;
import java.net.*;
import javax.swing.text.DefaultStyledDocument;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;
import javax.xml.transform.dom.*;

public class MissionEditor extends JFrame implements MouseListener, ActionListener, WindowListener, ListSelectionListener
{
    public static final String PREFERENCES_FILE_NAME = ".missionEditor.prefs";
    public static final String RESOURCE_PATH = "TheGame.jar";
    public static final String GENERIC_MONSTER_KEY = "(Generic)";
    
    private static final int COMPLETE_VICTORY = 0;
    private static final int VICTORY = 1;
    private static final int FAILURE = 2;
    private static final int COMPLETE_FAILURE = 3;
    
    private String [] monsterTypes = { "Marine", "Bob", "Brute", "BomberBrute", "ShotgunBrute", "CyberBrute", "Egg", "Scratcher", "Ripper", "Slasher", "SentryGun", "Oid", "SmartOid", "AlarmDrone", "AssaultDrone", "BattleRobot", "BrainRobot", "DestroyerRobot" };
    private String [] objectiveTypes = { "AddSomeMinesObjective", "AirlockDoorsObjective", "AssassinateObjective", "HackComputerObjective", "CloneVatObjective", "LockedDoorObjective", "RescueCiviliansObjective", "RecoverObjective", "RecoverDataObjective", "ReleaseLavaObjective", "ReturnToBaseObjective", "RipperSpawnObjective", "RobotFactoryObjective", "SabotageCommObjective", "SurvivalObjective", "BomberBruteParty" };

    private ClassLoader resourceLoader = null;
    private String missionFilePath = null;
    private File missionFile = null;
    
    private Vector players;
    private Vector monsters;
    private Vector objectives;
    private Objective primaryObjective = null;
    private Objective secondaryObjective = null;
    private HashMap randomBounds;
    private String missionName;
    private String longMissionDescription;
    private final String [] missionDebriefNames = { "Complete victory", "Victory", "Failure", "Complete failure" };
    private String [] missionDebriefDescriptions = { "Complete victory", "Victory", "Failure", "Complete failure" };
    private String [] nextMissionFileNames = { "", "", "", "" };
    private boolean [] canChangeMissionTeamAfter = { true, true, true, true };
    private int [] extraExperienceAfter = { 9, 6, 4, 2 };
        
    private int nextId=0;
    private EditorMap map = null;
    private String mapFileName;
    private MissionMapPanel mapPanel;
    private EditorOverviewMapPanel overview;
    private JMenuBar menubar;
    private JMenuItem save;
    private JMenuItem saveAs;
    private JMenuItem openMission;
    private JMenuItem openOldMission;
    private DirectionPanel directionPanel;
    private JComboBox monsterSelector;
    private CharacterInfoPanel characterInfoPanel;
    private JTabbedPane moSelector;
    private JList objectivesList;
    private ObjectiveListModel objectiveListModel;
    private JTextField objectiveClassName;
    private ObjectiveTableModel objectiveTableModel;
    private JTable objectiveTable;
    private JScrollPane jsp;
    private JComboBox objectiveTypeSelector;
    private JButton newObjectiveButton;
    private JButton deleteObjectiveButton;
    private JTextField numberOfObjectives;
    private JTextArea infoArea;
    private DefaultStyledDocument infoDocument;
    private JList missionEndConditionsList;
    private int lastSelectedMissionEndCondition = COMPLETE_VICTORY;
    private JTextArea missionDebriefArea;
    private JTextField nextMissionField;
    private JCheckBox canChangeMissionAfterBox;
    private JTextField extraExperienceAfterBox; 
    private JTextField missionNameField;
    private JTextArea longMissionDescriptionField;
    private JCheckBox endsOnStartBox;
    private JCheckBox hasPrequisiteBox;

    private int activeObjective = -1;

    public MissionEditor() 
    {
	super("MissionEditor");
	players = new Vector();
        monsters = new Vector();
        objectives = new Vector();
        randomBounds = new HashMap();
	objectiveTypeSelector = new JComboBox( objectiveTypes );	
	
        String directory = System.getProperty( "user.home" );
        try
        {
            File prefsFile = new File( directory, PREFERENCES_FILE_NAME );
            if( prefsFile.exists() )
            {
                System.out.println( "Loading preferences from " + prefsFile.toString() );
                BufferedReader prefsReader = new BufferedReader( new FileReader( prefsFile ) );
                missionFilePath = prefsReader.readLine();
                if( missionFilePath != null )
                {
                    missionFilePath = missionFilePath.trim();
		    loadPrefs();
                }
                else
                {
                    System.out.println( "Null path!" );
                }
                
            }
        }
        catch( Exception e )
        {
            System.out.println( "Exception loading preferences!" );
            e.printStackTrace();
        }
        URL resourceURL = null;
        try
        {
            resourceURL = new URL( RESOURCE_PATH );
        }
        catch( MalformedURLException mfue )
        {}
        URL [] urls = new URL[1];
        urls[0] = resourceURL;
        resourceLoader = new URLClassLoader( urls );
        
        // Open map for mission
	JFileChooser fileChooser = new JFileChooser( missionFilePath );
	SimpleFileFilter filter = new SimpleFileFilter( ".map" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Choose a map file" );
	int returnVal = fileChooser.showOpenDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    try
	    {
		File mapFile = fileChooser.getSelectedFile();
		mapFileName = mapFile.getName();
		map = EditorMap.load( mapFile );
                missionFilePath = mapFile.getParent();
                System.out.println( "Mission file path: " + missionFilePath );

	    }
            catch( NoClassDefFoundError ncdfe )
            {
                JOptionPane.showMessageDialog( this, "Resource file \"TheGame.jar\" was not found.\nMake sure it is in your class path.", "TheGame.jar not found", JOptionPane.ERROR_MESSAGE );
                System.exit( 1 );
            }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
		System.exit( 1 );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
		System.exit( 1 );
	    }
	    catch( Exception e )
	    {
		e.printStackTrace();
	    }
	}
	if( map != null )
	{	
	    // Go ahead...
	    Box base = new Box( BoxLayout.X_AXIS );
	    Box left = new Box( BoxLayout.Y_AXIS );
	    mapPanel = new MissionMapPanel( this, 48, 15, 10 );
	    mapPanel.addMouseListener( this );
	    JScrollPane scroller = new JScrollPane( mapPanel );
	    left.add( scroller );
            infoArea = new JTextArea();
            infoDocument = new DefaultStyledDocument();
            infoArea.setDocument( infoDocument );
            infoArea.setRows( 8 );
            infoArea.setTabSize( 8 );
            infoArea.setWrapStyleWord( true );
            JScrollPane infoScroller = new JScrollPane( infoArea );
            left.add( infoScroller );
	    base.add( left );
	    Box right = new Box( BoxLayout.Y_AXIS );
	    overview = new EditorOverviewMapPanel( this );
	    overview.addMouseListener( this );
	    Box overviewPanel = new Box( BoxLayout.Y_AXIS );
	    overviewPanel.add( overview );
	    overviewPanel.setBackground( Color.darkGray );
	    right.add( overviewPanel );
            // Create the table for Objectives
            moSelector = new JTabbedPane();
            Box objBox = Box.createVerticalBox();
	    objectiveListModel = new ObjectiveListModel( this );
	    objectivesList = new JList( objectiveListModel );
	    objectivesList.setVisibleRowCount( 5 );
	    objectivesList.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	    objectivesList.addListSelectionListener( this );
	    JScrollPane objectiveListPane = new JScrollPane( objectivesList, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER );
	    objBox.add( objectiveListPane );
            JPanel objCPanel = new JPanel();
            objectiveClassName = new JTextField( 25 );
            objCPanel.add( objectiveClassName );
            objBox.add( objCPanel );
            objectiveTableModel = new ObjectiveTableModel();
            objectiveTableModel.setMissionEditor( this );
            objectiveTable = new JTable( objectiveTableModel );
	    jsp = new JScrollPane( objectiveTable, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER );
	    objectiveTable.setRowSelectionAllowed( true );
	    objectiveTable.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	    objBox.add( jsp );
            JPanel oButtons = new JPanel();
	    oButtons.add( objectiveTypeSelector );
	    newObjectiveButton = new JButton( new ImageIcon( getClass().getResource( "PlusSign.png" ) ) );
            newObjectiveButton.addActionListener( this );
            oButtons.add( newObjectiveButton );
            deleteObjectiveButton = new JButton( new ImageIcon( getClass().getResource( "MinusSign.png" ) ) );
            deleteObjectiveButton.addActionListener( this );
            oButtons.add( deleteObjectiveButton );
	    objBox.add( oButtons );
	    objBox.add( Box.createVerticalGlue() );
            // Create panels for characters
            Box mBox = Box.createVerticalBox();            
            characterInfoPanel = new CharacterInfoPanel( this );
            mBox.add( characterInfoPanel );
            monsterSelector = new JComboBox( monsterTypes );
            mBox.add( monsterSelector );
            directionPanel = new DirectionPanel( 100, 100 );
            directionPanel.addActionListener( this );
            mBox.add( directionPanel );
	    mBox.add( Box.createVerticalGlue() );
            // Create panel for mission end conditions
            Box condBox = Box.createVerticalBox();
	    missionEndConditionsList = new JList( missionDebriefNames );
	    missionEndConditionsList.addListSelectionListener( this );
	    missionEndConditionsList.setSelectionMode( ListSelectionModel.SINGLE_SELECTION );
	    condBox.add( missionEndConditionsList );
	    condBox.add( Box.createVerticalStrut( 10 ) );
	    Box dBox = Box.createHorizontalBox();
	    JLabel debriefLabel = new JLabel( "Debrief" );
	    debriefLabel.setVerticalAlignment( SwingConstants.TOP );
	    dBox.add( debriefLabel );
	    dBox.add( Box.createHorizontalStrut( 10 ) );
	    missionDebriefArea = new JTextArea( 10, 25 );
	    dBox.add( new JScrollPane( missionDebriefArea ) );
	    condBox.add( dBox );
	    condBox.add( Box.createVerticalStrut( 10 ) );
	    Box nmBox = Box.createHorizontalBox();
	    nmBox.add( new JLabel( "Next mission" ) );
	    nmBox.add( Box.createHorizontalStrut( 10 ) );
	    nextMissionField = new JTextField();
	    nmBox.add( nextMissionField );
	    nmBox.add( Box.createHorizontalGlue() );
	    condBox.add( nmBox );
	    condBox.add( Box.createVerticalStrut( 10 ) );
	    canChangeMissionAfterBox = new JCheckBox( "Can change mission team afterwards", false );
	    condBox.add( canChangeMissionAfterBox );
	    Box eBox = Box.createHorizontalBox();
	    eBox.add( new JLabel( "Extra experience after" ) );
	    eBox.add( Box.createHorizontalStrut( 10 ) );
	    extraExperienceAfterBox = new JTextField( String.valueOf( extraExperienceAfter[ VICTORY ] ), 5 );
	    extraExperienceAfterBox.setHorizontalAlignment( SwingConstants.RIGHT );
	    eBox.add( extraExperienceAfterBox );
	    eBox.add( Box.createHorizontalGlue() );
	    condBox.add( eBox );
	    condBox.add( Box.createVerticalGlue() );	    
	    // Panel for general mission settings
	    Box genBox = Box.createVerticalBox();
            Box nBox = Box.createHorizontalBox();
            nBox.add( new JLabel( "Mission name" ) );
	    nBox.add( Box.createHorizontalStrut( 10 ) );
            missionNameField = new JTextField( 25 );
            missionNameField.setText( map.getName() );
            nBox.add( missionNameField );
	    nBox.add( Box.createHorizontalGlue() );
            genBox.add( nBox );
	    genBox.add( Box.createVerticalStrut( 10 ) );
            nBox = Box.createHorizontalBox();
            nBox.add( new JLabel( "Description" ) );
	    nBox.add( Box.createHorizontalStrut( 10 ) );
            longMissionDescriptionField = new JTextArea( 10, 25 );
            JScrollPane descScroller = new JScrollPane( longMissionDescriptionField );
            nBox.add( descScroller );
            genBox.add( nBox );
	    genBox.add( Box.createVerticalStrut( 10 ) );
	    endsOnStartBox = new JCheckBox( "Mission ends on start location", true );
	    genBox.add( endsOnStartBox );
	    genBox.add( Box.createVerticalStrut( 10 ) );	    
	    hasPrequisiteBox = new JCheckBox( "Mission has a prequisite", false );
	    genBox.add( hasPrequisiteBox );
	    genBox.add( Box.createVerticalGlue() );
               
	    moSelector.add( "General", genBox );
            moSelector.add( "Characters", mBox );
            moSelector.add( "Objectives", objBox );
            moSelector.add( "Finishing", condBox );

            right.add( moSelector );
            
	    base.add( right );
	    getContentPane().add( base );
	    
	    int shortCutMask = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
	    menubar = new JMenuBar();
	    JMenu file = new JMenu( "File" );
            openMission = new JMenuItem( "Open mission... " );
	    openMission.addActionListener( this );
	    file.add( openMission );
	    openOldMission = new JMenuItem( "Open old mission... " );
	    openOldMission.addActionListener( this );
	    file.add( openOldMission );
	    save = new JMenuItem( "Save" );
	    save.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_S, shortCutMask ) );
	    save.addActionListener( this );
	    file.add( save );
	    saveAs = new JMenuItem( "Save as... " );
	    saveAs.addActionListener( this );
	    file.add( saveAs );
	    menubar.add( file );
	    setJMenuBar( menubar );
            addWindowListener( this );
	    
	    pack();
	    SwingUtilities.invokeLater( 
		new Runnable()
		{
		    public void run()
		    {
			Dimension dimension = monsterSelector.getPreferredSize();
			monsterSelector.setMaximumSize( dimension );
			monsterSelector.invalidate();
			
			dimension = missionNameField.getPreferredSize();
			missionNameField.setMaximumSize( dimension );
			missionNameField.invalidate();
			
			dimension = extraExperienceAfterBox.getPreferredSize();
			extraExperienceAfterBox.setMaximumSize( dimension );
			dimension.width = missionNameField.getPreferredSize().width;
			nextMissionField.setMaximumSize( dimension );
			nextMissionField.setPreferredSize( dimension );
			
			javax.swing.table.TableColumn column;
			column = objectiveTable.getColumnModel().getColumn( 0 );
			column.setPreferredWidth( 50 );
			column = objectiveTable.getColumnModel().getColumn( 1 );
			column.setPreferredWidth( 100 );

			
			dimension = objectiveTable.getPreferredScrollableViewportSize();
			dimension.width = (int)(dimension.width*0.7);
			//objectiveTable.setPreferredSize( dimension );
			//objectiveTable.invalidate();
			objectiveTable.doLayout();
			//dimension.height = objectivesList.getPreferredSize().height;
			//objectivesList.setPreferredSize( dimension );
			
			validate();
			
			showFinishConditions( lastSelectedMissionEndCondition, false );
			setVisible( true );
		    }
		}
					);
	}
	else
	{
	    System.exit( 0 );
	}
    }
    
    public void addPlayerStartPoint( Point point )
    {
	if( map.isPassable( point ) )
	{
            CharacterReference marine = new CharacterReference( "Marine" );
            marine.setLocation( new Point( point ) );
            marine.setDirection( directionPanel.getDirection() );
            mapPanel.addSpriteForCharacter( marine );
	    players.add( marine );
            characterInfoPanel.setCharacter( marine );
	}
    }

    public void addMonster( Point point, String className )
    {
        if( map.isPassable( point ) )
        {
            CharacterReference monster = new CharacterReference( className );
            monster.setLocation( new Point( point ) );
            monster.setDirection( directionPanel.getDirection() );
	    monster.setId( String.valueOf( nextId++ ) );
            mapPanel.addSpriteForCharacter( monster );
            monsters.add( monster );
            characterInfoPanel.setCharacter( monster );
        }
            
    }
    
    public String addObjective( Objective objective )
    {
	if( objective.getId() == null )
	{
	    objective.setId( String.valueOf( objectives.size() ) );
	}
	objectives.add( objective );
	return objective.getId();
    }
    
    public void centerOnCharacter( TheGame.Character character )
    {
        mapPanel.centerOnCharacter( character );
        mapPanel.repaint();
    }

    public int getCharacterIndex( TheGame.Character character )
    {
        int index = -1;
        // Check the player characters
        index = players.indexOf( character );
        if( index > -1 )
            return index;
        // If that fails, check the monsters
        index = monsters.indexOf( character );

        return index;
    }
    
    public int getNumberOfPlayerStartPoints()
    {
	return players.size();
    }

    public Vector getPlayerCharacters()
    {
        return players;
    }
    
    public CharacterReference getPlayerCharacter( int index )
    {
        if( index < 0 || index >= players.size() )
            return null;
	return (CharacterReference)players.elementAt( index );
    }

    public Vector getMonsters()
    {
        return monsters;
    }

    public CharacterReference getMonster( int index )
    {
        if( index < 0 || index >= monsters.size() )
            return null;
        return (CharacterReference)monsters.elementAt( index );
    }

    public int getNumberOfObjectives()
    {
        return objectives.size();
    }

    public Objective getPrimaryObjective()
    {
	return primaryObjective;
    }
    
    public Objective getSecondaryObjective()
    {
	return secondaryObjective;
    }
    
    public Objective getObjective( int i )
    {
	if( i < 0 || i >= objectives.size() )
	{
	    return null;
	}
        return (Objective)objectives.elementAt( i );
    }

    public Objective getObjectiveById( String id )
    {
	if( id.equalsIgnoreCase( "Primary" ) )
	{
	    return primaryObjective;
	}
	if( id.equalsIgnoreCase( "Secondary" ) )
	{
	    return secondaryObjective;
	}
	try
	{
	    for( int i=0; i < objectives.size(); i++ )
	    {
		String theId = ((Objective)objectives.elementAt( i )).getId();
		if( theId != null && theId.equals( id ) )
		{
		    return (Objective)objectives.elementAt( i );
		}
	    }
	}
	catch( NumberFormatException nfe )
	{}
	System.err.println( "No match for objective id: " + id );
	return null;
    }    
    
    public boolean isPlayerCharacter( TheGame.Character character )
    {
        return players.contains( character );
    }
    
    public void loadPrefs()
    {
	File prefsFile = new File( missionFilePath + File.separator + "Resources", "MissionEditorPrefs.xml" );
	TheGame.ResourceManager.addBaseURL( "file://" + missionFilePath + File.separator + "Resources" + File.separator );
	System.out.println( "Trying to load prefs from: " + prefsFile.getAbsolutePath() );
	try
	{
	    if( prefsFile.exists() && prefsFile.canRead() )
	    {
		DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
		DOMImplementation builder = documentBuilder.getDOMImplementation();
	    	    
		System.out.println( "Parsing prefs file: " + prefsFile.getName() );
		org.w3c.dom.Document document = documentBuilder.parse( prefsFile );
	    	    
		org.w3c.dom.Element root = document.getDocumentElement();
		org.w3c.dom.Element monstersElement = XMLUtils.getNamedChild( root, "monsters" );
		org.w3c.dom.Element [] monsterElements = XMLUtils.getNamedChildren( monstersElement, "monster" );
		monsterTypes = new String[ monsterElements.length ];
		for( int i=0; i < monsterTypes.length; i++ )
		{
		    monsterTypes[i] = monsterElements[i].getAttribute( "classname" );
		}
		org.w3c.dom.Element objectivesElement = XMLUtils.getNamedChild( root, "objectives" );
		org.w3c.dom.Element [] objectiveElements = XMLUtils.getNamedChildren( objectivesElement, "objective" );
		objectiveTypes = new String[ objectiveElements.length ];
		objectiveTypeSelector.removeAllItems();
		for( int i=0; i < objectiveElements.length; i++ )
		{
		    objectiveTypes[i] = objectiveElements[i].getAttribute( "classname" );
		    objectiveTypeSelector.addItem( objectiveTypes[i] );
		}
		System.out.println( "Finished loading preferences" );
	    }
	}
	catch( Exception e )
	{
	    System.out.println( "Error loading preferences!" );
	    e.printStackTrace();
	}
    }
    
    public void mouseClicked( MouseEvent e )
    {
	if( e.getSource() == mapPanel )
	{
            if( SwingUtilities.isRightMouseButton( e ) )
            {
                if( monsterSelector.getSelectedIndex() == 0 )
		{
                    addPlayerStartPoint( mapPanel.getCell( e.getPoint() ) );
		}
                else
		{
                    addMonster( mapPanel.getCell( e.getPoint() ), (String)monsterSelector.getSelectedItem() );
		}
            }
            else
            {
                showCellInfo( mapPanel.getCell( e.getPoint() ) );
            }
	    mapPanel.repaint();
	    overview.repaint();
	}
	if( e.getSource() == overview )
	{
	    Point point = e.getPoint();
	    int div = overview.getCellSize();
	    Point cell = new Point( point.x/div, point.y/div );
            int cellSize = mapPanel.getCellSize();
	    mapPanel.scrollRectToVisible( new Rectangle( (cell.x-6)*cellSize, (cell.y-6)*cellSize, 13*cellSize, 13*cellSize ) );	    
	}
    }
    
    public void actionPerformed( ActionEvent e )
    {
        if( e.getSource() == newObjectiveButton )
        {
            Objective objective = null;
            String objectiveName = (String)objectiveTypeSelector.getSelectedItem();
            if( objectiveName == null )
            {
                return;
            }
            try {
		String prefix = "";
		if( !objectiveName.startsWith( "TheGame." ) )
		{
		    prefix = "TheGame.";
		}
                objective = (Objective)( Class.forName( prefix+objectiveName, true, resourceLoader ) ).newInstance();

                // Check whether we should set this as primary, secondary, or other objective
		int index = objectivesList.getSelectedIndex();
		
		if( index < 0 || index > 1  )
		{
		    addObjective( objective );
		}
		if( index == 0 )
		{
		    // Primary objective selected
		    if( primaryObjective == null )
		    {
			primaryObjective = objective;
			primaryObjective.setId( "Primary" );
		    }
		    else
		    {
			objectives.add( objective );
		    }
		}
		if( index == 1 )
		{
		    // Secondary objective selected
		    if( secondaryObjective == null )
		    {
			secondaryObjective = objective;
			secondaryObjective.setId( "Secondary" );
		    }
		    else
		    {
			addObjective( objective );
		    }
		}
		
		setVisibleObjective( objective );
		
		//objectivesList.setModel( objectiveListModel );
		objectiveListModel.update();
            }
            catch( Exception exception )
            {
                showError( exception.toString() );
            }
        }
        if( e.getSource() == deleteObjectiveButton )
        {
	    int index = objectivesList.getSelectedIndex();
	    if( index < 0 )
	    {
		Toolkit.getDefaultToolkit().beep();
	    }
	    if( index == 0 )
	    {
		// Try to delete primary objective
		if( primaryObjective != null )
		{
		    primaryObjective = null;
		    setVisibleObjective( null );
		}
		else
		{
		    Toolkit.getDefaultToolkit().beep();
		}
	    }
	    if( index == 1 )
	    {
		// Tro to delete secondary objective
		if( secondaryObjective != null )
		{
		    secondaryObjective = null;
		    setVisibleObjective( null );
		}
		else
		{
		    Toolkit.getDefaultToolkit().beep();
		}
	    }
	    if( index > 1 )
	    {
		if( objectives.size() > 0 )
		{
		    // Magic number: The actual objective indices are at index-2 because 0 and 1 are
		    // taken by primary and secondary objectives
		    objectives.removeElementAt( index-2 );
		    int newIndex = 0;
		    if( (index-2) < objectives.size() )
		    {
			newIndex = index-2;
		    }
		    else
		    {
			newIndex = objectives.size()-1;
		    }
		    Objective objective = getObjective( newIndex );
		    objectiveListModel.update();
		    setVisibleObjective( objective );
		}
                else
                {
                    Toolkit.getDefaultToolkit().beep();
                }
            }
	    objectiveListModel.update();
        }
        if( e.getSource() == directionPanel )
        {
            if( ( e.getModifiers() & ActionEvent.META_MASK ) == ActionEvent.META_MASK )
            {
                characterInfoPanel.setCharacterDirection( directionPanel.getDirection() );
            }
        }
	if( e.getSource() == save )
	{
	    saveMission();
	}
	if( e.getSource() == saveAs )
	{
	    saveMissionWithFileChooser();
	}
	if( e.getSource() == openMission )
	{
	    openMissionWithFileChooser( false );
	    showFinishConditions( VICTORY, false );
	}
	if( e.getSource() == openOldMission )
	{
	    openMissionWithFileChooser( true );
	    showFinishConditions( VICTORY, false );
	}
	
    }

    public void removeMonster( int index )
    {
        if( index >= 0 && index < monsters.size() )
        {
            CharacterReference character = (CharacterReference)monsters.elementAt( index );
            monsters.removeElementAt( index );
            mapPanel.removeSpriteForCharacter( character );
            mapPanel.repaint();
        }
    }
    
    public void removePlayerCharacter( int index )
    {
        if( index >= 0 && index < players.size() )
        {
            CharacterReference character = (CharacterReference)players.elementAt( index );
            players.removeElementAt( index );
            mapPanel.removeSpriteForCharacter( character );
            mapPanel.repaint();
        }
    }

    public void setRandomBoundsForCharacterClass( int minimum, int maximum )
    {
        String className = (String)monsterSelector.getSelectedItem();
        setRandomBoundsForCharacterClass( className, minimum, maximum );
    }
    
    public void setRandomBoundsForCharacterClass( String className, int minimum, int maximum )
    {
        Point bounds;
        int min = minimum;
        int max = maximum;
        if( min < 0 )
        {
            min = 0;
        }
        if( max < 0 )
        {
            max = 0;
        }
        if( min > max )
        {
            max = min;
        }
        if( randomBounds.containsKey( className ) )
        {
            bounds = (Point)randomBounds.get( className );
            bounds.x = min;
            bounds.y = max;
        }
        else
        {
            bounds = new Point( min, max );
            randomBounds.put( className, bounds );
        }
    }

    public int getRandomBoundsMinimumForCharacterClass( String className )
    {
        int min = 0;
        if( randomBounds.containsKey( className ) )
        {
            Point bounds = (Point)randomBounds.get( className );
            min = bounds.x;
        }
        return min;
    }
    
    public int getRandomBoundsMaximumForCharacterClass( String className )
    {
        int max = 0;
        if( randomBounds.containsKey( className ) )
        {
            Point bounds = (Point)randomBounds.get( className );
            max = bounds.y;
        }
        return max;
    }
    
    public void setVisibleObjective( Objective objective )
    {
	objectiveTableModel.setObjective( objective );
	if( objective != null )
	{
	    objectiveClassName.setText( objective.getClassName() );
	    objectiveTable.doLayout();
	    objectiveTable.invalidate();
	    jsp.invalidate();
	    objectiveTable.validate();
	    validate();
        }
	else
	{
	    objectiveClassName.setText( "Not set" );
	}
    }

    public void openMissionWithFileChooser( boolean old )
    {
        JFileChooser fileChooser = new JFileChooser( missionFilePath );
        SimpleFileFilter filter = new SimpleFileFilter( ".mission" );
        fileChooser.setFileFilter( filter );
        fileChooser.setDialogTitle( "Load mission" );
        int returnVal = fileChooser.showOpenDialog( this );
        if( returnVal == JFileChooser.APPROVE_OPTION )
        {
            try
            {
		if( old )
		{
		    openOldMissionFromFile( fileChooser.getSelectedFile() );
		}
		else
		{
		    missionFile = fileChooser.getSelectedFile();
		    openMissionFromFile( missionFile );
		}
                setTitle( "Mapmaker: " + fileChooser.getSelectedFile().getName() );
		objectiveListModel.update();
            }
            catch( FileNotFoundException e )
            {
                System.out.println( "File not found!" );
            }
            catch( IOException e )
            {
                System.out.println( e.toString() );
            }
	    catch( Exception e )
	    {
		e.printStackTrace();
	    }
        }
    }
    
    public void saveMission() 
    {
	if( missionFile == null )
	{
	    saveMissionWithFileChooser();
	}
	else
	{
	    try
	    {
		// Ensure the end conditions are set
		showFinishConditions( lastSelectedMissionEndCondition, true );
		// Save
		saveMissionToFile( missionFile );
		setTitle( "Mission Editor: " + missionFile.getName() );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
	    catch( Exception e )
	    {
		e.printStackTrace();
	    }
	}
    }
    
    public void saveMissionWithFileChooser()
    {
	JFileChooser fileChooser = new JFileChooser();
	SimpleFileFilter filter = new SimpleFileFilter( ".mission" );
	fileChooser.setFileFilter( filter );
	fileChooser.setDialogTitle( "Save mission" );
	int returnVal = fileChooser.showSaveDialog( this );
	if( returnVal == JFileChooser.APPROVE_OPTION )
	{
	    if( fileChooser.getSelectedFile().exists() )
	    {
		int selection = JOptionPane.showConfirmDialog( this, "Are you sure you want to overwrite the file?", "", JOptionPane.YES_NO_CANCEL_OPTION );
                if( selection == JOptionPane.CANCEL_OPTION || selection == JOptionPane.NO_OPTION || selection == JOptionPane.CLOSED_OPTION )
                {
                    return;
                }
	    }
	    try
	    {
		// Ensure the end conditions are set
		showFinishConditions( lastSelectedMissionEndCondition, true );
		// Save
		saveMissionToFile( fileChooser.getSelectedFile() );
		setTitle( "Mission Editor: " + fileChooser.getSelectedFile().getName() );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "File not found!" );
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
	    }
	    catch( Exception e )
	    {
		e.printStackTrace();
	    }
	}
    }

    public void openMissionFromFile( File file ) throws Exception
    {
	if( file.exists() && file.canRead() )
	{
	    DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	    DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
	    DOMImplementation builder = documentBuilder.getDOMImplementation();
	    
	    FileInputStream fis = new FileInputStream( file );
	    GZIPInputStream gis = new GZIPInputStream( fis );
	    
	    System.out.println( "Parsing mission file: " + file.getName() );
	    Document document = documentBuilder.parse( gis );
	    gis.close();
	    fis.close();
	    
	    Element root = document.getDocumentElement();

	    missionName = root.getAttribute( "name" );
	    missionNameField.setText( missionName );
	    boolean hasPrequisite = new Boolean( root.getAttribute( "hasprequisite" ) ).booleanValue();
	    hasPrequisiteBox.setSelected( hasPrequisite );
	    boolean endsAtBase = new Boolean( root.getAttribute( "endsatbase" ) ).booleanValue();
	    endsOnStartBox.setSelected( endsAtBase );

	    mapFileName = root.getAttribute( "mapfile" );
	    try
	    {
		File mapFile = new File( file.getParent(), mapFileName );
		map = EditorMap.load( mapFile );
		mapPanel.setMap( map );
	    }
	    catch( FileNotFoundException e )
	    {
		System.out.println( "Map file not found!" );
		return;
	    }
	    catch( IOException e )
	    {
		System.out.println( e.toString() );
		System.exit( 1 );
	    }
	    catch( Exception e )
	    {
		e.printStackTrace();
		System.exit( 1 );
	    }
	    // Load player start points
	    Element playersElement = XMLUtils.getNamedChild( root, "players" );
	    if( playersElement != null )
	    {
		Element [] playerElements = XMLUtils.getNamedChildren( playersElement, "startpoint" );
		int x, y, dir;
		for( int i = 0; i < playerElements.length; i++ )
		{
		    x = Integer.parseInt( playerElements[i].getAttribute( "x" ) );
		    y = Integer.parseInt( playerElements[i].getAttribute( "y" ) );
		    dir = Integer.parseInt( playerElements[i].getAttribute( "direction" ) );
		    CharacterReference marine = new CharacterReference( "Marine" );
		    marine.setLocation( new Point( x, y ) );
		    marine.setDirection( dir );
		    mapPanel.addSpriteForCharacter( marine );
		    players.add( marine );
		    characterInfoPanel.setCharacter( marine );
		}
	    }
	    // Load monsters
	    Element monstersElement = XMLUtils.getNamedChild( root, "monsters" );
	    if( monstersElement != null )
	    {
		/// Load actual monsters
		monsters = new Vector();
		CharacterReference monster;
		String monsterClass;
		int x, y, dir;
		Element [] monsterElements = XMLUtils.getNamedChildren( monstersElement, "monster" );
		for( int i = 0; i < monsterElements.length; i++ )
		{
		    monsterClass = monsterElements[i].getAttribute( "class" );
		    x = Integer.parseInt( monsterElements[i].getAttribute( "x" ) );
		    y = Integer.parseInt( monsterElements[i].getAttribute( "y" ) );
		    dir = Integer.parseInt( monsterElements[i].getAttribute( "direction" ) );
		    monster = new CharacterReference( monsterClass );
		    monster.setLocation( new Point( x, y ) );
		    monster.setDirection( dir );
		    String id = monsterElements[i].getAttribute( "id" );
		    if( id == null || id.trim().length() < 1 )
		    {
			id = String.valueOf( i );
		    }
		    monster.setId( id );
		    String monsterName = monsterElements[i].getAttribute( "name" );
		    if( monsterName != null )
		    {
			monster.setName( monsterName );
		    }
		    String monsterCustomHP = monsterElements[i].getAttribute( "customhitpoints" );
		    if( monsterCustomHP != null )
		    {
			monster.setCustomHitPoints( monsterCustomHP );
		    }
		    String monsterCustomAP = monsterElements[i].getAttribute( "customactionpoints" );
		    if( monsterCustomAP != null )
		    {
			monster.setCustomActionPoints( monsterCustomAP );
		    }
		    monster.setDormant( new Boolean( monsterElements[i].getAttribute( "dormant" ) ).booleanValue() );

		    mapPanel.addSpriteForCharacter( monster );
		    monsters.add( monster );
		    characterInfoPanel.setCharacter( monster );
		}
		/// Load random bounds
		randomBounds = new HashMap();
		Element [] randomBoundsElements = XMLUtils.getNamedChildren( monstersElement, "randommonsterbounds" );
		for( int i=0; i < randomBoundsElements.length; i++ )
		{
		    String aKey = randomBoundsElements[i].getAttribute( "class" );
		    int min = Integer.parseInt( randomBoundsElements[i].getAttribute( "minimum" ) );
		    int max = Integer.parseInt( randomBoundsElements[i].getAttribute( "maximum" ) );
		    setRandomBoundsForCharacterClass( aKey, min, max );
		}
	    }
	    // Load the Objectives
	    Element objectivesElement = XMLUtils.getNamedChild( root, "objectives" );
	    if( objectivesElement != null )
	    {
		Element primaryObjectiveElement = XMLUtils.getNamedChild( objectivesElement, "primaryobjective" );
		if( primaryObjectiveElement != null )
		{
		    primaryObjective = Objective.decodeFromElement( XMLUtils.getNamedChild( primaryObjectiveElement, "objective" ) );
		}
		else
		{
		    primaryObjective = null;
		}
		Element secondaryObjectiveElement = XMLUtils.getNamedChild( objectivesElement, "secondaryobjective" );
		if( secondaryObjectiveElement != null )
		{
		    secondaryObjective = Objective.decodeFromElement( XMLUtils.getNamedChild( secondaryObjectiveElement, "objective" ) );
		}
		else
		{
		    secondaryObjective = null;
		}
		
		objectives = new Vector();
		Element [] objectiveElements = XMLUtils.getNamedChildren( objectivesElement, "objective" );
		for( int i=0; i < objectiveElements.length; i++ )
		{
		    Objective objective = Objective.decodeFromElement( objectiveElements[i] );
		    objectives.add( objective );
		    setVisibleObjective( getObjective( 0 ) );
		}
	    }
	    // Load mission end conditions and descriptions
	    Element longDescriptionElement = XMLUtils.getNamedChild( root, "description" );
	    if( longDescriptionElement != null )
	    {
		longMissionDescription = longDescriptionElement.getAttribute( "description" );
		longMissionDescriptionField.setText( longMissionDescription );
	    }
	    Element conditionsElement = XMLUtils.getNamedChild( root, "endconditions" );
	    if( conditionsElement != null )
	    {
		Element [] endConditionElements = XMLUtils.getNamedChildren( conditionsElement, "victorycondition" );
		for( int i=0; i < endConditionElements.length; i++ )
		{
		    decodeCondition( endConditionElements[i] );
		}
	    }
	}
	    
    }
    
    public void openOldMissionFromFile( File file ) throws FileNotFoundException, IOException
    {
        FileInputStream fis = new FileInputStream( file );
        DataInputStream reader = new DataInputStream( fis );
        mapFileName = reader.readUTF();
        try
        {
            File mapFile = new File( file.getParent(), mapFileName );
            map = EditorMap.load( mapFile );
            mapPanel.setMap( map );
        }
        catch( FileNotFoundException e )
        {
            System.out.println( "Map file not found!" );
	    return;
        }
        catch( IOException e )
        {
            System.out.println( e.toString() );
            System.exit( 1 );
        }
	catch( Exception e )
	{
	    e.printStackTrace();
	    System.exit( 1 );
	}
        // Read player start points
        players = new Vector();
        int playerStartPoints = reader.readInt();
        int x, y, dir;
        for( int i = 0; i < playerStartPoints; i++ )
        {
            x=reader.readInt();
            y=reader.readInt();
            dir=reader.readInt();
            CharacterReference marine = new CharacterReference( "Marine" );
            marine.setLocation( new Point( x, y ) );
            marine.setDirection( dir );
            mapPanel.addSpriteForCharacter( marine );
            players.add( marine );
            characterInfoPanel.setCharacter( marine );
        }
        // Read Objectives
        objectives = new Vector();
        int objNum = reader.readInt();
	String objId = null;
        for( int i=0; i < objNum; i++ )
        {
	    try
	    {
		Objective objective = Objective.load( reader );
		objId = addObjective( objective );
		setVisibleObjective( objective );
	    }
	    catch( IllegalArgumentException iae )
	    {}
        }
        // Read monsters
        monsters = new Vector();
        CharacterReference monster;
        String monsterClass;
        int monsterNum = reader.readInt();
        for( int i = 0; i < monsterNum; i++ )
        {
            monsterClass = reader.readUTF();
            x = reader.readInt();
            y = reader.readInt();
            dir = reader.readInt();
            monster = new CharacterReference( monsterClass );
            monster.setLocation( new Point( x, y ) );
            monster.setDirection( dir );
            mapPanel.addSpriteForCharacter( monster );
	    monster.setId( String.valueOf( i ) );
            monsters.add( monster );
            characterInfoPanel.setCharacter( monster );
        }
        // Read character random bounds
        randomBounds = new HashMap();
        int randomNum = reader.readInt();
        for( int i=0; i < randomNum; i++ )
        {
            String aKey = reader.readUTF();
            int min = reader.readInt();
            int max = reader.readInt();
            setRandomBoundsForCharacterClass( aKey, min, max );
        }
        // Read mission descriptions
        missionName = reader.readUTF();
        missionNameField.setText( missionName );
        longMissionDescription = reader.readUTF();
        longMissionDescriptionField.setText( longMissionDescription );
        String missionCompleteDescription = reader.readUTF();
        missionDebriefDescriptions[ VICTORY ] = missionCompleteDescription;
        String missionFailDescription = reader.readUTF();
        missionDebriefDescriptions[ FAILURE ] = missionFailDescription;
	lastSelectedMissionEndCondition = -1;
        reader.close();
        fis.close();
	overview.repaint();
	primaryObjective = null;
	secondaryObjective = null;
    }
    
    public void saveMissionToFile( File file ) throws Exception
    {
	DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
	DocumentBuilder documentBuilder = documentBuilderFactory.newDocumentBuilder();
	DOMImplementation builder = documentBuilder.getDOMImplementation();
	
	org.w3c.dom.Document document = builder.createDocument( null, "mission", null );
	Element root = document.getDocumentElement();
	missionName = missionNameField.getText();
	if( missionName.trim().length() < 1 )
	{
	    missionName = map.getName();
	}
	root.setAttribute( "name", missionName );
	root.setAttribute( "mapfile", mapFileName );
	root.setAttribute( "hasprequisite", String.valueOf( hasPrequisiteBox.isSelected() ) );
	root.setAttribute( "endsatbase", String.valueOf( endsOnStartBox.isSelected() ) );

	// Store player start points
	Element playerStartPointsElement = document.createElement( "players" );
	for( int i=0; i < getNumberOfPlayerStartPoints(); i++ )
	{
	    Element startPointElement = document.createElement( "startpoint" );
	    startPointElement.setAttribute( "x", String.valueOf( getPlayerCharacter(i).getLocation().x ) );
	    startPointElement.setAttribute( "y", String.valueOf( getPlayerCharacter(i).getLocation().y ) );
            startPointElement.setAttribute( "direction", String.valueOf( getPlayerCharacter(i).getDirection() ) );
	    playerStartPointsElement.appendChild( document.createTextNode( "\n" ) );
	    playerStartPointsElement.appendChild( startPointElement ); 						  
	}
	root.appendChild( document.createTextNode( "\n" ) );
	root.appendChild( playerStartPointsElement );
        // store monsters
	Element monstersElement = document.createElement( "monsters" );
	for( int i=0; i < monsters.size(); i++ )
        {
	    Element monsterElement = document.createElement( "monster" );
            CharacterReference monster = getMonster( i );
            monsterElement.setAttribute( "class", monster.getClassName() );
            monsterElement.setAttribute( "x", String.valueOf( monster.getLocation().x ) );
            monsterElement.setAttribute( "y", String.valueOf( monster.getLocation().y ) );
            monsterElement.setAttribute( "direction", String.valueOf( monster.getDirection() ) );
	    monsterElement.setAttribute( "id", monster.getId() );
	    monsterElement.setAttribute( "dormant", String.valueOf( monster.isDormant() ) );
	    String monsterName = monster.getName();
	    if( monsterName != null && monsterName.trim().length() > 0 )
	    {
		monsterElement.setAttribute( "name", monsterName );
	    }
	    String monsterCustomHitPoints = monster.getCustomHitPoints();
	    if( monsterCustomHitPoints != null )
	    {
		monsterElement.setAttribute( "customhitpoints", monsterCustomHitPoints );
	    }
	    String monsterCustomActionPoints = monster.getCustomActionPoints();
	    if( monsterCustomActionPoints != null )
	    {
		monsterElement.setAttribute( "customactionpoints", monsterCustomActionPoints );
	    }	    
	    monstersElement.appendChild( document.createTextNode( "\n" ) );
	    monstersElement.appendChild( monsterElement );
        }
        // Save monster random bounds
        Set boundKeys = randomBounds.keySet();
        Object [] theKeys = boundKeys.toArray();
	
        for( int i=0; i < theKeys.length; i++ )
        {
            String aKey = (String)theKeys[i];
	    Element randomMonstersElement = document.createElement( "randommonsterbounds" );
	    randomMonstersElement.setAttribute( "class", aKey );
	    randomMonstersElement.setAttribute( "minimum", String.valueOf( getRandomBoundsMinimumForCharacterClass( aKey ) ) );
            randomMonstersElement.setAttribute( "maximum", String.valueOf( getRandomBoundsMaximumForCharacterClass( aKey ) ) );
	    monstersElement.appendChild( document.createTextNode( "\n" ) );
	    monstersElement.appendChild( randomMonstersElement );
        }
	root.appendChild( document.createTextNode( "\n" ) );
	root.appendChild( monstersElement );
        // Store the Objectives
	Element objectivesElement = document.createElement( "objectives" );
	if( primaryObjective != null )
	{
	    Element primaryObjectiveElement = document.createElement( "primaryobjective" );
	    primaryObjectiveElement.appendChild( primaryObjective.encodeForDocument( document ) );
	    objectivesElement.appendChild( document.createTextNode( "\n" ) );
	    objectivesElement.appendChild( primaryObjectiveElement );
	}
	if( secondaryObjective != null )
	{
	    Element secondaryObjectiveElement = document.createElement( "secondaryobjective" );
	    secondaryObjectiveElement.appendChild( secondaryObjective.encodeForDocument( document ) );
	    objectivesElement.appendChild( document.createTextNode( "\n" ) );
	    objectivesElement.appendChild( secondaryObjectiveElement );
	}
        for( int i=0; i < objectives.size(); i++ )
        {
            Objective objective = (Objective)objectives.elementAt( i );
	    objectivesElement.appendChild( document.createTextNode( "\n" ) );
	    objectivesElement.appendChild( objective.encodeForDocument( document ) );
        }
	root.appendChild( document.createTextNode( "\n" ) );
	root.appendChild( objectivesElement );
	// Store mission end conditions and descriptions
	longMissionDescription = longMissionDescriptionField.getText().trim();
	if( longMissionDescription.length() > 0 )
	{
	    Element longDescriptionElement = document.createElement( "description" );
	    longDescriptionElement.setAttribute( "description", longMissionDescription );
	    root.appendChild( document.createTextNode( "\n" ) );
	    root.appendChild( longDescriptionElement );
	}
	Element conditionsElement = document.createElement( "endconditions" );
	if( primaryObjective != null )
	{
	    Element victoryElement = encodeCondition( document, VICTORY );
	    conditionsElement.appendChild( document.createTextNode( "\n" ) );
	    conditionsElement.appendChild( victoryElement );
	    Element failureElement = encodeCondition( document, FAILURE );
	    conditionsElement.appendChild( document.createTextNode( "\n" ) );
	    conditionsElement.appendChild( failureElement );
	}
	if( secondaryObjective != null )
	{
	    Element completeVictoryElement = encodeCondition( document, COMPLETE_VICTORY );
	    conditionsElement.appendChild( document.createTextNode( "\n" ) );
	    conditionsElement.appendChild( completeVictoryElement );
	    Element completeFailureElement = encodeCondition( document, COMPLETE_FAILURE );
	    conditionsElement.appendChild( document.createTextNode( "\n" ) );
	    conditionsElement.appendChild( completeFailureElement );
	}
	root.appendChild( document.createTextNode( "\n" ) );
	root.appendChild( conditionsElement );
			  
	System.out.println( "Saving mission" );
	
	//System.out.println( document.toString() );
	
	DOMSource domSource = new DOMSource( document );
		
	FileOutputStream fos = new FileOutputStream( file );
	GZIPOutputStream gos = new GZIPOutputStream( fos );
	StreamResult streamResult = new StreamResult( gos );
        //StreamResult streamResult = new StreamResult( file );
        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        Transformer transformer = transformerFactory.newTransformer();
        transformer.transform( domSource, streamResult );        
	
	gos.close();
	fos.close();
	System.out.println( "Mission saved" );
	
    }
    
    private void decodeCondition( Element victoryElement )
    {
	String conditionName = victoryElement.getAttribute( "condition" );
	int condition = 0;
	for( int i = 0; i < missionDebriefNames.length; i++ )
	{
	    if( conditionName.equals( missionDebriefNames[ i ] ) )
	    {
		condition = i;
	    }
	}
	String nextMissionName = victoryElement.getAttribute( "nextmission" );
	if( nextMissionName.equalsIgnoreCase( "Null" ) )
	{
	    nextMissionName = null;
	}
	nextMissionFileNames[ condition ] = nextMissionName;
	canChangeMissionTeamAfter[ condition ] = new Boolean( victoryElement.getAttribute( "canchangeteamafter" ) ).booleanValue();
	extraExperienceAfter[ condition ] = Integer.parseInt( victoryElement.getAttribute( "extraexperienceafter" ) );
	    
	String debrief = victoryElement.getAttribute( "debrief" );
	missionDebriefDescriptions[ condition ] = debrief;
    }
    
    private Element encodeCondition( org.w3c.dom.Document document, int condition )
    {
	Element victoryElement = document.createElement( "victorycondition" );
	victoryElement.setAttribute( "condition", missionDebriefNames[ condition ] );
	String nextMissionName = nextMissionFileNames[ condition ];
	if( nextMissionName == null )
	{
	    nextMissionName = "Null";
	}
	victoryElement.setAttribute( "nextmission", nextMissionName );
	victoryElement.setAttribute( "canchangeteamafter", String.valueOf( canChangeMissionTeamAfter[ condition ] ) );
	victoryElement.setAttribute( "extraexperienceafter", String.valueOf( extraExperienceAfter[ condition ] ) );
	String debrief = missionDebriefDescriptions[ condition ];
	victoryElement.setAttribute( "debrief", debrief );
	return victoryElement;
    }
    
    public void saveOldMissionToFile( File file ) throws FileNotFoundException, IOException
    {
	FileOutputStream fos = new FileOutputStream( file );
	DataOutputStream writer = new DataOutputStream( fos );
	writer.writeUTF( mapFileName );
        // Save the player start points
	writer.writeInt( getNumberOfPlayerStartPoints() );
	for( int i=0; i < getNumberOfPlayerStartPoints(); i++ )
	{
	    writer.writeInt( getPlayerCharacter(i).getLocation().x );
	    writer.writeInt( getPlayerCharacter(i).getLocation().y );
            writer.writeInt( getPlayerCharacter(i).getDirection() );
	}
        // Save the Objectives
        writer.writeInt( objectives.size() );
        for( int i=0; i < objectives.size(); i++ )
        {
            Objective objective = (Objective)objectives.elementAt( i );
            objective.save( writer );
        }
        // Save monsters
        writer.writeInt( monsters.size() );
        for( int i=0; i < monsters.size(); i++ )
        {
            CharacterReference monster = getMonster( i );
            writer.writeUTF( monster.getClassName() );
            writer.writeInt( monster.getLocation().x );
            writer.writeInt( monster.getLocation().y );
            writer.writeInt( monster.getDirection() );
        }
        // Save monster random bounds
        Set boundKeys = randomBounds.keySet();
        Object [] theKeys = boundKeys.toArray();
        writer.writeInt( theKeys.length );
        for( int i=0; i < theKeys.length; i++ )
        {
            String aKey = (String)theKeys[i];
            writer.writeUTF( aKey );
            writer.writeInt( getRandomBoundsMinimumForCharacterClass( aKey ) );
            writer.writeInt( getRandomBoundsMaximumForCharacterClass( aKey ) );
        }
        // Save mission descriptions
        missionName = missionNameField.getText();
        writer.writeUTF( missionName );
        longMissionDescription = longMissionDescriptionField.getText();
        writer.writeUTF( longMissionDescription );
        String missionCompleteDescription = missionDebriefDescriptions[ VICTORY ];
        writer.writeUTF( missionCompleteDescription );
        String missionFailDescription = missionDebriefDescriptions[ FAILURE ];
        writer.writeUTF( missionFailDescription );
	writer.flush();
	writer.close();
	fos.close();
    }

    public void savePreferences()
    {
        String directory = System.getProperty( "user.home" );
        try
        {
            File prefsFile = new File( directory, PREFERENCES_FILE_NAME );
            if( !prefsFile.exists() )
            {
                prefsFile.createNewFile();
            }
            System.out.println( "Saving preferences to " + prefsFile.toString() );
            BufferedWriter prefsWriter = new BufferedWriter( new FileWriter( prefsFile ) );
            if( missionFilePath != null )
            {
                prefsWriter.write( missionFilePath );
                prefsWriter.newLine();
            }
            prefsWriter.flush();
            prefsWriter.close();
            
        }
        catch( Exception e )
        {
            System.out.println( "Exception savong preferences!" );
            e.printStackTrace();
        }
        
    }
    
    public static String pointToString( Point point )
    {
        String pointString = "("+point.x+", "+point.y+")";
        return pointString;
    }

    public static Point pointFromString( String string ) throws IllegalArgumentException
    {
        StringTokenizer st = new StringTokenizer( string.trim(), "(, )", true );
        int x, y;
        Point thePoint = null;
        boolean fubar = false;	// Is there a mistake in parsing the string
        int tokens = st.countTokens();
        if( tokens == 5 || tokens == 6 )
        {
            try
            {
                st.nextToken(); 			// Remove the first parenthesis
                x = Integer.parseInt( st.nextToken() ); // Read x
                st.nextToken();				// Remove the comma
                if( tokens == 6 )
                {
                    st.nextToken();			// Remove the space
                }
                y = Integer.parseInt( st.nextToken() );	// Read y
                thePoint = new Point( x, y );
            }
            catch( NumberFormatException e )
            {
                fubar = true;
            }
        }
        else
        {
            fubar = true;
        }
        if( fubar )
        {
            throw new IllegalArgumentException( "Malformed point: " + string );
        }
        return thePoint;
    }

    public void showCellInfo( Point point )
    {
        if( map.getCell( point ) != null )
        {
	    Objective currentObjective = null;
	    Class selectionClass = null;
	    String selectedPropertyName = null;
	    boolean failed = false;
	    if( objectiveTable.getSelectedRow() > -1 )
	    {
		currentObjective = objectiveTableModel.getObjective();
		if( currentObjective != null )
		{
		    selectedPropertyName = (String)(objectiveTableModel.getValueAt( objectiveTable.getSelectedRow(), 0 ));
		    selectionClass = currentObjective.getPropertyClass( selectedPropertyName );
		    System.out.println( "Property class: " + selectionClass );
		    
		    if( selectionClass != null )
		    {
			if( selectionClass.equals( TheGame.Door.class ) )
			{
			    if( map.getDoor( point.x, point.y ) != null )
			    {
				currentObjective.setProperty( selectedPropertyName, Objective.pointToString( point ) );
			    }
			    else
			    {
				failed = true;
			    }
			}
			if( selectionClass.equals( Point.class ) )
			{
			    currentObjective.setProperty( selectedPropertyName, Objective.pointToString( point ) );
			}
			if( selectionClass.equals( TheGame.Monster.class ) )
			{
			    CharacterReference monster;
			    for( int i=0; i < monsters.size(); i++ )
			    {
				monster = getMonster( i );
				if( monster.getLocation().equals( point ) )
				{
				    currentObjective.setProperty( selectedPropertyName, monster.getId() );
				}
			    }
			}
			if( selectionClass.equals( TheGame.MapObject.class ) )
			{
			    TheGame.MapObject mapObject = map.getObjectForPoint( point );
			    if( mapObject != null )
			    {
				currentObjective.setProperty( selectedPropertyName, mapObject.getId() );
			    }
			}
		    }
		}
	    }
	    if( !failed )
	    {
		objectiveTable.clearSelection();
		objectiveTable.invalidate();
		validate();
		objectiveTable.repaint();
	    }
	    else
	    {
		Toolkit.getDefaultToolkit().beep();
	    }
	    infoArea.setText( "" );
            infoArea.append( "Point:  \t " + Objective.pointToString( point ) + "\n");
            if( map.getDoor( point.x, point.y ) != null )
            {
                infoArea.append( "Has door" + "\n" );
            }
	    TheGame.Room room = map.getRoomForPoint( point );
	    if( room != null )
	    {
		infoArea.append( "Room: " + room.toString() + "\n" );
	    }
	    TheGame.MapObject mapObject = map.getObjectForPoint( point );
	    if( mapObject != null )
	    {
		infoArea.append( "Object: " + mapObject.getName() + " (" + mapObject.getId() + ") \n" );
	    }
            for( int i=0; i < players.size(); i++ )
            {
                if( getPlayerCharacter( i ).getLocation().equals( point ) )
                {
                    infoArea.append( "Player: \t " + i + "\n" );
		    characterInfoPanel.setCharacter( getPlayerCharacter( i ) );
                }
            }
            for( int i=0; i < monsters.size(); i++ )
            {
                CharacterReference monster;
                monster = getMonster( i );
                if( monster.getLocation().equals( point ) )
                {
		    String infoLine = "Monster:\t " + monster.getId() + " (" + monster.getType();
		    if( monster.getName() != null )
		    {
			infoLine = infoLine + " called " + monster.getName();
		    }
		    infoLine = infoLine + ")\n";
                    infoArea.append( infoLine );
		    characterInfoPanel.setCharacter( monster );
                }
            }
            for( int i=0; i < objectives.size(); i++ )
            {
                Objective objective;
                objective = getObjective( i );
                try
                {
                    if( objective.isLocatedIn( point ) )
                    {
                        infoArea.append( "Objective:\t " + objective.getId() + " (" + objective.getClass().getName() + ")\n" );
                    }
                }
                catch( Exception e )
                {}
            }
	    if( primaryObjective != null )
	    {
		if( primaryObjective.isLocatedIn( point ) )
		{
		    infoArea.append( "Primary Objective:\t (" + primaryObjective.getClass().getName() + ")\n" );
		}
		
	    }
	    if( secondaryObjective != null )
	    {
		if( secondaryObjective.isLocatedIn( point ) )
		{
		    infoArea.append( "Secondary Objective:\t (" + secondaryObjective.getClass().getName() + ")\n" );
		}
		
	    }
            infoArea.setCaretPosition( infoArea.getText().length() );
        }
    }
    
    public void showError( String error )
    {
        //infoArea.setForeground( Color.red );	// Pain... This sets the colour for the whole view
        System.out.println( error );
        infoArea.append( error+"\n" );
        Toolkit.getDefaultToolkit().beep();
        //infoArea.setForeground( Color.black );
    }
    
    public void showFinishConditions( int condition, boolean update )
    {
	// Store the old description
	if( update )
	{
	    missionDebriefDescriptions[ lastSelectedMissionEndCondition ] = missionDebriefArea.getText();
	    nextMissionFileNames[ lastSelectedMissionEndCondition ] = nextMissionField.getText();
	    canChangeMissionTeamAfter[ lastSelectedMissionEndCondition ] = canChangeMissionAfterBox.isSelected();
	    int value = 0;
	    try
	    {
		value = Integer.parseInt( extraExperienceAfterBox.getText() );
	    }
	    catch( NumberFormatException nfe )
	    {
		value = 0;
	    }
	    extraExperienceAfter[ lastSelectedMissionEndCondition ] = value;
	}
	// Replace with new
	lastSelectedMissionEndCondition = condition;
	missionDebriefArea.setText( missionDebriefDescriptions[ lastSelectedMissionEndCondition ] );
	nextMissionField.setText( nextMissionFileNames[ lastSelectedMissionEndCondition ] );
	canChangeMissionAfterBox.setSelected( canChangeMissionTeamAfter[ lastSelectedMissionEndCondition ] );
	extraExperienceAfterBox.setText( String.valueOf( extraExperienceAfter[ lastSelectedMissionEndCondition ] ) );
	if( !update )
	{
	    missionEndConditionsList.setSelectedIndex( condition );
	}
    }
    
    public void mousePressed( MouseEvent e )
    {}
    
    public void mouseReleased( MouseEvent e )
    {}
    
    public void mouseEntered( MouseEvent e )
    {}
    
    public void mouseExited( MouseEvent e )
    {}

    public void valueChanged( ListSelectionEvent e )
    {
	if( e.getSource() == objectivesList )
	{
	    Objective objective = null;
	    int index = objectivesList.getSelectedIndex();
	    if( index == 0 )
	    {
		objective = primaryObjective;
	    }
	    if( index == 1 )
	    {
		objective = secondaryObjective;
	    }
	    if( index > 1 )
	    {
		objective = getObjective( index-2 );
	    }
	    setVisibleObjective( objective );
	}
	if( e.getSource() == missionEndConditionsList )
	{
	    int index = missionEndConditionsList.getSelectedIndex();
	    showFinishConditions( index, true );
	}
    }
    
    public void windowClosed( WindowEvent e )
    {
        savePreferences();
        System.exit( 0 );
    }

    public void windowClosing( WindowEvent e )
    {
        savePreferences();
        System.exit( 0 );
    }

    public void windowActivated( WindowEvent e )
    {}
    
    public void windowDeactivated( WindowEvent e )
    {}

    public void windowOpened( WindowEvent e )
    {}

    public void windowIconified( WindowEvent e )
    {}

    public void windowDeiconified( WindowEvent e )
    {}
    
    public EditorMap getMap()
    {
	return map;
    }
    
    public static void main(String args[]) 
    {
	System.out.println( "Starting" );
	new MissionEditor();
    }
}

class ObjectiveListModel extends AbstractListModel
{
    private MissionEditor editor = null;
    
    public ObjectiveListModel( MissionEditor editor )
    {
	this.editor = editor;
    }
    
    public Object getElementAt( int index )
    {
	//System.out.println( "Getting List element at: " + index );
	// Magic numbers: We are using 0 for primary objective, 1 for secondary, and 2+ for normal
	// objectives (whose indices start at 0. Why do you think this isn't a public class? ;) )
	if( index == 0 )
	{
	    String name = "";
	    if( editor.getPrimaryObjective() != null )
	    {
		name = ": " + editor.getPrimaryObjective().getClassName();
	    }
	    return "Primary objective" + name;
	}
	if( index == 1 )
	{
	    String name = "";
	    if( editor.getSecondaryObjective() != null )
	    {
		name = ": " + editor.getSecondaryObjective().getClassName();
	    }
	    return "Secondary objective" + name;
	}
	return editor.getObjective( index-2 ).getId() + ": " + editor.getObjective( index-2 ).getClassName();
    }
    
    public int getSize()
    {
	return editor.getNumberOfObjectives()+2;
    }
    
    public void update()
    {
	fireContentsChanged( this, 0, editor.getNumberOfObjectives()+2 );
    }
}
