// Objective.java
// Started on July 11th, 2001
// Copyright (c) Virtu Halttunen

package TheGame;

import java.awt.Point;
import java.util.StringTokenizer;
import java.io.*;
import org.w3c.dom.*;

public abstract class Objective
{
    public static final String LOCATION_KEY = "Location";
    public static final String SUCCESS_MESSAGE = "Success message";
    public static final String IS_SHOWN_ON_MAP_KEY = "Show on map";
    
    protected Game activeGame = null;
    protected String id = null;
 
    public Game getGame()
    {
	return activeGame;
    }
    
    public void setGame( Game game )
    {
	activeGame = game;
    }
          
    public String getId()
    {
	return id;
    }
    
    public void setId( String id )
    {
	this.id = id;
    }
    
    public int getNumberOfSprites()
    {
	return 0;
    }
    
    public Sprite getSprite( int x, int y )
    {
	return null;
    }
    
    public String getClassName()
    {
	String className = getClass().getName();
	if( className.startsWith( "TheGame." ) )
	{
	    className = className.substring( 8 );
	}
	return className;
    }

    public abstract String [] getPropertyNames();

    public abstract String getProperty( String key );
    
    public Class getPropertyClass( String key )
    {
	if( key.equals( Objective.LOCATION_KEY ) )
	{
	    return Point.class;
	}
	if( key.equals( Objective.IS_SHOWN_ON_MAP_KEY ) )
	{
	    return Boolean.class;   
	}
	
	return String.class;
    }

    public abstract void setProperty( String key, String value ) throws IllegalArgumentException;

    public void setLocation( Point point )
    {}
    
    public void setLocation( int x, int y )
    {
	setLocation( new Point( x, y ) );
    }
    
    public String getDescriptionClip()
    {
	return "";
    }
    
    public static Objective decodeFromElement( Element element )
    {
	Objective objective = null;
	if( element.getTagName().equals( "objective" ) )
	{
	    try
	    {
		String className = element.getAttribute( "class" );
		if( !className.startsWith( "TheGame." ) )
		{
		    className = "TheGame."+className;
		}
		objective = (Objective)( Class.forName( className ).newInstance() );
		objective.setId( element.getAttribute( "id" ) );
		Element [] propertyElements = XMLUtils.getNamedChildren( element, "property" );
		if( propertyElements != null )
		{
		    for( int i=0; i < propertyElements.length; i++ )
		    {
			String propertyName = propertyElements[i].getAttribute( "key" );
			String propertyValue = propertyElements[i].getAttribute( "value" );
			if( propertyValue.equalsIgnoreCase( "null" ) )
			{
			    propertyValue = null;
			}
			objective.setProperty( propertyName, propertyValue );
		    }
		}
		//System.out.println( "Loaded Objective: " + objective.getClassName() );
	    }
	    catch( Exception ex )
	    {
		System.out.println( "Exception loading objective!" );
		ex.printStackTrace();
	    }
	}
	return objective;
    }
    
    public static Objective load( DataInputStream inStream )
    {
        int loadedKeys = 0;
        String loadedName;
        String loadedKeyName;
        String loadedValue;
        Objective objective = null;
        
        try
        {
            loadedName = inStream.readUTF();
	    if( !loadedName.startsWith( "TheGame." ) )
	    {
		loadedName = "TheGame."+loadedName;
	    }
            objective = (Objective)( Class.forName( loadedName ).newInstance() );
            loadedKeys = inStream.readInt();
            for( int i=0; i < loadedKeys; i++ )
            {
                loadedKeyName = inStream.readUTF();
                loadedValue = inStream.readUTF();
                if( loadedValue.equalsIgnoreCase( "Null" ) )
                {
                    loadedValue = null;
                }
		try
		{
		    objective.setProperty( loadedKeyName, loadedValue );
		}
		catch( IllegalArgumentException iae )
		{
		    System.out.println( "Illegal argument exception!" );
		    System.out.println( iae.toString() );
		}
            }
        }
        catch( Exception e )
        {
            System.out.println( "Error loading Objective:" );
            e.printStackTrace();
            System.exit( 0 );
        }
        return objective;
    }

    public Element encodeForDocument( Document document )
    {
	Element objectiveElement = document.createElement( "objective" );
        objectiveElement.setAttribute( "class", getClassName() );
	objectiveElement.setAttribute( "id", id );
        String savedValue = null;
        String [] savedProperties = getPropertyNames();
        if( savedProperties != null )
        {
	    for( int i=0; i < savedProperties.length; i++ )
	    {
		savedValue = getProperty( savedProperties[i] );
		if( savedValue == null )
		{
		    savedValue = "Null";
		}
		Element propertyElement = document.createElement( "property" );
		propertyElement.setAttribute( "key", savedProperties[i] );
		propertyElement.setAttribute( "value", savedValue );
		objectiveElement.appendChild( document.createTextNode( "\n" ) );
		objectiveElement.appendChild( propertyElement );
	    }
	}
	return objectiveElement;
    }
    
    public void save( DataOutputStream outStream ) throws IOException
    {
        int savedKeys = 0;
        String savedName;
        String savedKeyName;
        String savedValue;
        String [] savedProperties = getPropertyNames();
        if( savedProperties != null )
        {
            savedKeys = savedProperties.length;
        }
        savedName = getClass().getName();
        outStream.writeUTF( savedName );
        outStream.writeInt( savedKeys );
        for( int i=0; i < savedKeys; i++ )
        {
            savedKeyName = savedProperties[ i ];
            outStream.writeUTF( savedKeyName );
            savedValue = getProperty( savedKeyName );
            if( savedValue == null )
            {
                savedValue = "Null";
            }
            outStream.writeUTF( savedValue );
        }
    }
    
    /**
     *	This method is called once for each objective at the end of the turn.
     *	It can cause the Objective to do something.
     *
     *	If the Objective is the primary objective, returning true means the
     *	mission can end. For non-primary objectives, the return value is ignored.
     *
     *	@return true if the mission can be ended
     */
    public boolean isCompleted()
    {
	return true;
    }
    
    /**
     *	Is the objective successful. Only matters on primary and secondary objective.
     *
     *	@return true if successful
     */
    public boolean isSuccessful()
    {
	return true;
    }
    
    public boolean isLocatedIn( Point point )
    {
	return false;
    }
    
    public boolean isLocatedIn( int x, int y )
    {
	return false;
    }
        
    public boolean isFailed()
    {
	return false;
    }
    
    public boolean isUsable()
    {
	return false;
    }
    
    public boolean isRemoteControlled()
    {
	return false;
    }

    /**
     * Returns the size of the objective. The size varies from 0 to 100
     * A character can inhabit the same square with the Objective only if the sum
     * of their respective sizes is less than 100. Size also affects ranged
     * weapon attacks. If the Objective's size is greater than the weapon's
     * blockingSize, the Objective blocks the line of fire for that weapon.
     *
     * @return 		the size of the Objective
     */
    
    public int getSize()
    {
        return 0;
    }
    
    public boolean isDestructible()
    {
	return false;
    }
    
    public boolean isShownOnMap()
    {
        return false;
    }
    
    // Returns true if destroyed
    public boolean handleDamage( Point point, int damage )
    {
	return false;
    }
    
    /*
     * Ask the Objective to handle the particular action
     *
     * @param action	the action to be handled
     * @return		<code>true</code> if meaningful for this Objective
     *
     */
    public boolean handleAction( GameAction action )
    {
        return false;
    }

    public static boolean booleanFromString( String string ) throws IllegalArgumentException
    {
        if( string.trim().equalsIgnoreCase( "True" ) )
        {
            return true;
        }
        if( string.trim().equalsIgnoreCase( "False" ) )
        {
            return false;
        }
        throw new IllegalArgumentException( "Not a boolean: " + string );
    }

    public static String booleanToString( boolean aBool )
    {
        if( aBool )
        {
            return "True";
        }
        return "False";
    }
    
    public static double doubleFromString( String string ) throws IllegalArgumentException
    {
        double number = 0;
        try
        {
            number = Double.parseDouble( string );
        }
        catch( NumberFormatException nfe )
        {
            throw new IllegalArgumentException( "Not a number: " + string );
        }
        return number;
    }
    
    
    public static int intFromString( String string ) throws IllegalArgumentException
    {
        int number = 0;
        try
        {
            number = Integer.parseInt( string );
        }
        catch( NumberFormatException nfe )
        {
            throw new IllegalArgumentException( "Not a number: " + string );
        }
        return number;
    }
    
    public static String [] parseStrings( String string )
    {
	String [] tokens;
	//System.out.println( "Objective parsing strings from: " + string );
	StringTokenizer tokenizer = new StringTokenizer( string, "{}, ", false );
	tokens = new String[ tokenizer.countTokens() ];
	for( int i=0; i < tokens.length; i++ )
	{
	    tokens[i] = tokenizer.nextToken();
	    //System.out.println( i + ": " + tokens[i] );
	}
	return tokens;
    }

    public static String pointToString( Point point )
    {
	if( point == null )
	{
	    return null;
	}
        String pointString = "("+point.x+", "+point.y+")";
        return pointString;
    }

    public static Point pointFromString( String string ) throws IllegalArgumentException
    {
	if( string == null )
	{
	    throw new IllegalArgumentException( "Null string" );
	}
        StringTokenizer st = new StringTokenizer( string.trim(), "(, )", true );
        int x, y;
        Point thePoint = null;
        boolean fubar = false;	// Is there a mistake in parsing the string
        int tokens = st.countTokens();
        if( tokens == 5 || tokens == 6 )
        {
            try
	    {
		st.nextToken(); 			// Remove the first parenthesis
		x = Integer.parseInt( st.nextToken() ); // Read x
		st.nextToken();				// Remove the comma
		if( tokens == 6 )
		{
		    st.nextToken();			// Remove the space
		}
		y = Integer.parseInt( st.nextToken() );	// Read y
		thePoint = new Point( x, y );
	    }
            catch( NumberFormatException e )
	    {
                fubar = true;
	    }
        }
        else
        {
            fubar = true;
        }
        if( fubar )
        {
            throw new IllegalArgumentException( "Malformed point: " + string );
        }
        return thePoint;
    }
    
}