// Tribble.java
// Started on August 4th, 2005
// Copyright (c) Virtu Halttunen

package TheGame;

import java.awt.Point;
import java.util.Vector;

public class Tribble extends Monster
{
    public static final int ROAMING = 0;
    public static final int ESCAPING = 1;
    public static final int ATTACKING = 2;
    public static final int PANICKING = 3;
    
    private boolean hasMoreActions = false;
    Character target = null;
    private Point goal = null;
    private int state = ROAMING;
    private boolean pickedUp = false;

    public Tribble()
    {
        location = new Point( 0, 0 );
        direction = Map.NORTH;
        alive = true;
        lastAction = new WaitAction( this, location );
        maximumActionPoints = 5;
        currentActionPoints = 5;
        maximumHitPoints = 4;
        currentHitPoints = 4;
    }

    public boolean isPickedUp()
    {
	return pickedUp;
    }
    
    public void setPickedUp( boolean pickedUp )
    {
	this.pickedUp = pickedUp;
	dormant = pickedUp;
    }
    
    public void setSpottedPlayer( Character character, boolean canSee )
    {
	super.setSpottedPlayer( character, canSee );
        target = character;
        if( canSee )
        {
	    // Runs away from the players
	    if( goal == null )
	    {
		preAction();
	    }
	    GameAction action = playAction();
	    if( action instanceof MoveAction )
	    {
		MoveAction mAction = (MoveAction)action;
		direction = Map.getDirection( location, mAction.getEndPoint(), direction );
		activeGame.moveCharacter( this, direction );
		currentActionPoints++;
	    }
	}
    }

    private Character [] getVisiblePlayers()
    {
	Character [] visiblePlayers = null;
	Vector help = new Vector();
	for( int i=0; i < activeGame.getPlayerCharacters().size(); i++ )
	{
	    Character check = activeGame.getPlayerCharacter( i );
	    if( check.isAlive() )
	    {
		if( activeGame.isCellVisibleToCharacter( check, location ) )
		{
		    help.add( check );
		}
	    }
	}
	visiblePlayers = new Character[ help.size() ];
	for( int i=0; i < help.size(); i++ )
	{
	    visiblePlayers[i] = (Character)help.elementAt( i );
	}
	return visiblePlayers;
    }
    
    public void preAction()
    {
	Character [] visibleEnemies = getVisiblePlayers();
	if( visibleEnemies.length > 0 )
	{
	    goal = SeekerAI.getSafePointForCharacter( this, visibleEnemies );
            state = ESCAPING;
	}
    }
    
    public GameAction playAction()
    {
        Point cell;

        Monster character;
        Character[] enemies;

        if( alive )
        {
            if( currentActionPoints > 0 )
            {
                enemies = getVisiblePlayers();
                if( enemies.length > 0 && goal == null )
                {
		    goal = SeekerAI.getSafePointForCharacter( this, enemies );
                    state = ESCAPING;                   
                }
                if( state == ESCAPING )
                {
		    if( goal == null )
		    {
			state = PANICKING;
		    }
		    else
		    {
			Vector path = activeGame.getMap().findShortestPath( this, goal, false );
			if( path.size() >= 2 )	
			{
			    cell = ((MapNode)path.elementAt( 1 )).getPoint();
			    direction = Map.getDirection( location, cell, direction );
			    lastAction = new MoveAction( this, new Point( location ), new Point( cell ) );
			    return lastAction;
			}
			else
			{
			    // If found goal, start roaming
			    if( activeGame.getMap().isNeighbor( goal, location ) )
			    {
				state = ROAMING;
				goal = null;
			    }
			    else
			    {
				// We are probably blocked
				state = PANICKING;
			    }
			}
                    }
                }
		if( state == PANICKING )
		{
		    // Run away from the closest enemy
		    Character check;
		    Character target = this;
		    double range = Double.MAX_VALUE;
		    double checkRange;
		    int newDirection = -1;
		    int tempDirection = -1;
		    for( int i = 0; i < enemies.length; i++ )
		    {
			check = enemies[ i ];
			checkRange = location.distance( check.getLocation() );
			if( checkRange < range && check.isAlive() )
			{
			    target = check;
			    range = checkRange;
			}
		    }
		    tempDirection = Map.getDirection( location, target.getLocation(), direction );
		    // Check the opposite direction
		    newDirection = (tempDirection+4)%8;
		    if( activeGame.getMap().isPassable( location, newDirection ) )
		    {
			// This will do
			direction = newDirection;
		    }
		    else
		    {
			newDirection = (tempDirection+3)%8;
			if( activeGame.getMap().isPassable( location, newDirection ) )
			{
			    direction = newDirection;
			}
			else
			{
			    newDirection = (tempDirection+5)%8;
			    if( activeGame.getMap().isPassable( location, newDirection ) )
			    {
				direction = newDirection;
			    }
			    else
			    {
				newDirection = -1;
			    }
			}
		    }
		    // If found suitable escape direction, run there (add some randomness, so the Tribbles don't all end up in same spot
		    if( newDirection > -1 && Math.random() < 0.6 )
		    {
			System.out.println( "Tribble panicking from (" + location.x +","+location.y +") towards " +newDirection );
			lastAction = new MoveAction( this, new Point( location ), activeGame.getMap().getNeighbor( location, direction ));
			return lastAction;
		    }
		    else
		    {
			state = ROAMING;
		    }
		}
		
                if( state == ROAMING )
                {
                    // Move
                    if( activeGame.canCharacterMove( this, direction, false ) )
                    {
                        cell = activeGame.getMap().getNeighbor( location, direction );
                        lastAction = new MoveAction( this, new Point( location ), new Point( cell ) );
                        return lastAction;
                    }
                    direction = (int)(Math.random()*8);                    
                }
            }
            lastAction = new WaitAction( this, new Point( location ) );
        }
        return lastAction;
    }
    
    public void postAction()
    {
	Character [] visibleEnemies = activeGame.getVisibleEnemies( this );

	if( visibleEnemies.length < 1 )
	{
	    state = ROAMING;
	}
    }

    public GameAction getLastAction()
    {
        return lastAction;
    }
    
    public int getMeleeAbility()
    {
	return 1;
    }
    
    public int getMeleeCost()
    {
	return 1;
    }

    public void resetActionsPoints()
    {
        currentActionPoints = maximumActionPoints;
    }

    public boolean hasMoreActions()
    {
        if( currentActionPoints > 0 )
        {
            return true;
        }
        return false;
    }
    
    public boolean isFoe( Character character )
    {
	if( character.getType().equals( "Klingon" ) )
	{
	    return true;
	}
        return false;
    }    

    public boolean applyDamage( int damage )
    {
	currentHitPoints = currentHitPoints-damage;
        lastAction = new WoundAction( this, location );
        if( currentHitPoints <= 0 )
        {
            alive = false;
            lastAction = new DeadAction( this, new Point( location ) );
            maximumActionPoints = 0;
            currentActionPoints = 0;
	}
        return alive;
    }

    public int getSize()
    {
        return 10;
    }

}